/**
  *  \file game/global.cc
  *  \brief Global Turn Data
  *
  *  This contains most of the GGameTurn class.
  *
  *  (c) 2001-2006 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/global.h"
#include "game/ship.h"
#include "util/nls.h"
#include "game/player.h"
#include "util/console.h"
#include "util/except.h"
#include "game/planet.h"
#include "cpluslib/format.h"
#include "util/recode.h"
#include "game/host.h"

/** Global turn instance. */
Ptr<GGameTurn> current_turn_internal;
Ptr<GGameTurn> display_turn_internal;

/******************************* GGameTurn *******************************/

/*! \class GGameTurn
    This class contains everything that makes up a game turn: things
    that change each turn. There's one global instance of this class
    which is used by most code; some special-purpose code will work
    on different instances.

    This contains many functions that otherwise would be global,
    hence the huuuuuuuge interface.

    During a run, we expect all planet/starbase objects to be present,
    but possibly empty. To not have to do this during static
    initialisation (which isn't possible with Ptr<> anyway), one
    must call init() before loading the turn. Basically, do
    \code
        try {
            GGameTurn& t = getCurrentTurn();
            t.init();
            // load game...
            t.postprocess();
        }
        catch(std::exception& e) {
            // ...
        }
    \endcode
*/

/** Constructor. Creates a blank GGameTurn. You must call init()
    before you can use it. */
GGameTurn::GGameTurn()
    : playing_set(), data_set()
{
    std::fill_n(player_data, countof(player_data), (Player*) 0);
}

/** Destructor. Frees all memory. */
GGameTurn::~GGameTurn()
{
    done();
}

/** Create all data. */
void
GGameTurn::init()
{
    dat.init();
    dis.init();
    vcrs = new GClassicVcrDatabase(GClassicVcrDatabase::vcr_Host);
}

/** Clear all data. */
void
GGameTurn::done()
{
    dat.done();
    dis.done();
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        delete player_data[i-1];
        player_data[i-1] = 0;
    }
    vcrs = 0;
}

GGameTurn::Player&
GGameTurn::createPlayerData(int player)
{
    ASSERT(player > 0 && player <= NUM_PLAYERS);
    if (!player_data[player-1])
        player_data[player-1] = new Player(player);
    return *player_data[player-1];
}

const GGameTurn::Player&
GGameTurn::getPlayerDataNC(int player) const
{
    ASSERT(player > 0 && player <= NUM_PLAYERS);
    return *player_data[player-1];
}

GGameTurn::Player&
GGameTurn::getPlayerData(int player)
{
    // FIXME: not much difference to createPlayerData
    ASSERT(player > 0 && player <= NUM_PLAYERS);
    return player_data[player-1] ? *player_data[player-1] : createPlayerData(player);
}


/** Set GEN data for player.
    \param player player
    \param gen parsed GENx.DAT for that player. */
void
GGameTurn::setGen(int player, const GGen& gen)
{
    getPlayerData(player).gen = gen;
    if (!dat.getTurnNumber()) {
        dat.setTurnNumber(gen.getTurnNumber());
        dis.setTurnNumber(gen.getTurnNumber());
    }
}

/** Set that we have data for the specified player. */
void
GGameTurn::setHaveData(int player)
{
    createPlayerData(player);
    data_set |= player;
}

/** Load Ion Storm Names. */
void
GGameTurn::loadStormNames()
{
    Ptr<Stream> p = spec_file_dir->openFileNT("storm.nm", Stream::C_READ);
    if (!p) {
        /* dummy version */
        for (int i = 1; i <= NUM_STORMS; ++i) {
            dat.getIonStorm(i).setName(_("Noname Storm"));
            dis.getIonStorm(i).setName(_("Noname Storm"));
        }
    } else {
        char buffer[NUM_STORMS * 20];
        if (p->read(buffer, sizeof buffer) != sizeof buffer)
            throw FileTooShortException("storm.nm");
        Recoder::getE2I().recodeArray(buffer, sizeof buffer);
        for (int i = 1; i <= NUM_STORMS; ++i) {
            const string_t name = getBasicStringN(&buffer[20 * (i-1)], 20);
            dat.getIonStorm(i).setName(name);
            dis.getIonStorm(i).setName(name);
        }
    }
}

/** Load Planet Names. */
void
GGameTurn::loadPlanetNames()
{
    Ptr<Stream> p = spec_file_dir->openFileT("planet.nm", Stream::C_READ);
    char buffer[NUM_PLANETS * 20];
    if (p->read(buffer, sizeof buffer) != sizeof buffer)
        throw FileTooShortException("planet.nm");
    Recoder::getE2I().recodeArray(buffer, sizeof buffer);
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        const string_t name = getBasicStringN(&buffer[20 * (i-1)], 20);
        dat.getPlanet(i).setName(name);
        dis.getPlanet(i).setName(name);
    }
}

/** Load Planet Coordinates. */
void
GGameTurn::loadPlanetXY(Stream& s)
{
    char buffer[NUM_PLANETS * 6];
    if (s.read(buffer, sizeof buffer) != sizeof buffer)
        throw FileTooShortException(s.getName());
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        int x = getInt16(&buffer[6*i-6]);
        int y = getInt16(&buffer[6*i-4]);
        if (! GPlanet::isClippedCoord(GPoint(x, y))) {
            dat.getPlanet(i).setXY(x, y);
            dis.getPlanet(i).setXY(x, y);
        }
    }
}

/** Postprocess after loading. */
void
GGameTurn::postprocess()
{
    // FIXME: use real values; handle the case that this is not
    // the current turn and therefore not being played.
    playing_set = data_set;

    loadPlanetNames();
    loadStormNames();
    GHost::finish();

    console.write(LOG_INFO, _("Compiling starchart..."));

    dat.postprocess(playing_set);
    dis.postprocess(playing_set);

    chart_bbox.initFromUniverse(dat);

    // inbox.sort(); // FIXME: configurable

    console.write(LOG_INFO, _("=== Loading complete ==="));
    for (int i = 1; i <= NUM_PLAYERS; ++i)
        if (isPlaying(i))
            console.write(LOG_INFO, format(_("Playing player %d: %s"), i, getFullName(i)));
}

GGameTurn::Player::Player(int pid)
    : outbox(pid)
{ }

GGameTurn::Player::~Player()
{ }

/**************************** Global Functions ***************************/

Signal0 sig_turn_changed;

void
setDisplayedTurn(Ptr<GGameTurn> disp)
{
    GGameTurn* orig = display_turn_internal;
    if (disp)
        display_turn_internal = disp;
    else
        display_turn_internal = current_turn_internal;
    if (orig != display_turn_internal)
        sig_turn_changed.raise();
}

void
initTurn()
{
    current_turn_internal = new GGameTurn();
    setDisplayedTurn(0);
}

void
initGameGlobals(string_t gamedir, string_t rootdir)
{
    initDirectories(gamedir, rootdir);
    initTurn();
    GHost::clear();
}
