/**
  *  \file game/msg.cc
  *  \brief Subspace Message Handling
  *
  *  Implementation of the abstract GMessagebox class (any message box)
  *  and the standard GInbox.
  *
  *  (c) 2001-2006 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include "game/msg.h"
#include "game/struct.h"
#include "util/nls.h"
#include "util/except.h"
#include "game/pref.h"
#include "game/player.h"
#include "cpluslib/assert.h"
#include "util/recode.h"
#include "util/misc.h"

/****************************** GMessagebox ******************************/

/*! \class GMessagebox
    This class provides an abstract interface to message boxes; basically
    everything that's needed to implement a standard message reader.
    In addition, it maintains a current position. Messages are indexed
    0-based. */
GMessagebox::GMessagebox()
    : current(0)
{ }

GMessagebox::~GMessagebox()
{ }

/** Set current message. If the index is out of range, it is
    automatically adjusted. If the change results in a change of the
    currently-selected message, raises sig_sel_changed. */
void
GMessagebox::setCurrent(int aindex)
{
    int count = getCount();
    if (aindex >= count)
        aindex = count-1;
    if (aindex < 0)
        aindex = 0;
    if (aindex != current) {
        current = aindex;
        sig_sel_changed.raise();
    }
}

/********************************* GInbox ********************************/

/*! \class GInbox

    GInbox provides handling for incoming messages. It is able to read
    standard inbox files (RST and MDATA). */

/** Message structure. */
struct GInbox::Message {
    string_t text;
    int      status;

    Message(const string_t& str)
        : text(str), status(0) { }
};

/** Constructur. Creates empty messagebox. */
GInbox::GInbox()
{ }

/** Destructor. */
GInbox::~GInbox()
{
    safeClear(messages);
}

int
GInbox::getCount()
{
    return messages.size();
}

string_t
GInbox::getText(int index)
{
    ASSERT(index >= 0 && unsigned(index) < messages.size());
    return messages[index]->text;
}

static string_t
tweakHeader(string_t what)
{
    string_t::size_type n;
    while ((n = what.find_first_of("<>")) != string_t::npos)
        what.erase(n, 1);
    return strTrim(strLCWords(what));
}

/** Get Message Heading. This is the same algorithm as in PCC 1.x. */
string_t
GInbox::getHeading(int index)
{
    using namespace std;
    string_t line = strTrim(strFirst(getText(index), "\n"));
    if (line.length() < 5)
        /* translators: must start with "(  )" */
        return _("(  ) Unknown");
    if (line[0] != '(')
        /* pre-3.2 message format */
        return "(  ) " + tweakHeader(line);
    if (line[1] == 'o' && user_pref.CollapseOldMessages)
        /* translators: must start with "(oX)" */
        return _("(oX) Message from earlier turn");

    /* It is a message in our preferred format */
    string_t pre = line.substr(0, 3) + ") ";
    switch(toupper((unsigned char) pre[2])) {
     case 'R':
        {
            int i = getPlayerIdFromChar(pre[3]);
            if (i > 0)
                return pre + getFullName(i);
            else if (i == 0)
                return pre + _("Anonymous Message");
            else
                break;
        }
     case 'D':
        return pre + _("Starbase Message");
     case 'L':
        return pre + _("Minefield Laid");
     case 'I':
        return pre + _("Ion Storm");
     case 'G':
        return pre + _("HConfig");
    }
    strRemove(line, ")");
    return pre + tweakHeader(line);
}

int
GInbox::getStatus(int index)
{
    ASSERT(index >= 0 && unsigned(index) < messages.size());
    return messages[index]->status;
}

void
GInbox::setStatus(int index, int newstat)
{
    ASSERT(index >= 0 && unsigned(index) < messages.size());
    messages[index]->status = newstat;
}

/** Add a single message.
    \param str Complete text of message */
void
GInbox::addMessage(const string_t str)
{
    safeAppend(messages, new Message(str));
}

/** Sort messages. This will group messages of equal subjects
    together, but preserves the overall order (i.e. messages from
    other races will remain first). */
void
GInbox::sort()
{
    std::vector<Message*> new_data;
    std::vector<bool> did;
    did.resize(messages.size());
    new_data.reserve(messages.size());

    /* This algorithm is nominally O(n^2). */
    for (unsigned int i = 0; i < messages.size(); ++i) {
        if (!did[i]) {
            const string_t h = getHeading(i);

            // Copy over all messages with this subject
            for (unsigned int j = i; j < messages.size(); ++j) {
                if (!did[j] && getHeading(j) == h) {
                    did[j] = true;
                    new_data.push_back(messages[j]);
                }
            }
        }
    }
    messages.swap(new_data);
}

/** Load Inbox. This reads RST and MDATA files.
    \param str File. Pointer is set to beginning of message directory.
    File must be seekable.
    \throws FileFormatException on error. Caller should add file name. */
void
GInbox::loadInbox(Stream& str)
{
    /* read count */
    TInt16 count;
    if (!getStructure(str, count))
        throw FileFormatException(_("Message file too short"));
    if (count.value < 0)
        throw FileFormatException(_("Message file is invalid"));
    if (count.value == 0)
        return;

    /* read message directory */
    uint32 size = count.value * uint32(6);
    Array<char> dir(size);
    if (str.read(dir.ptr(), size) != size)
        throw FileFormatException(_("Message file too short"));

    int16 max_size = 0;
    for (int i = 0; i < count.value; ++i) {
        /* verify message directory entry */
        TIncomingMessageHeader hdr;
        getStructure(dir.ptr() + 6*i, hdr);
        if (hdr.address <= 0 || hdr.length <= 0)
            throw FileFormatException(_("Message file is invalid"));
        if (hdr.length > max_size)
            max_size = hdr.length;
    }

    /* load the messages */
    Array<char> buffer(max_size);
    for (int i = 0; i < count.value; ++i) {
        TIncomingMessageHeader hdr;
        getStructure(dir.ptr() + 6*i, hdr);
        if (!str.seek(hdr.address-1) || str.read(buffer.ptr(), hdr.length) != hdr.length)
            throw FileIOException(_("Unable to read message file"));
        addMessage(getMessageFromArray(buffer.ptr(), hdr.length, true));
    }
}

/******************************* Utilities *******************************/

/** Parse a char array into a message. This applies a few fixups to
    the message:
    - rot13 is decoded, of course;
    - Dominate fixup: drop everything from NUL to EOL;
    - Winplan fixup (if rewrap is on), see below.
    - Recoding character sets.

    The message is returned in C++ format (lines separated by '\n'),
    no matter what that means in binary.

    Winplan fixup: older Winplans send messages with CR+LF line
    terminators, not CR as usual. Host messes these up because it
    thinks the LFs are text characters, and inserts additional CRs.
    Therefore, we change the meaning of the control characters such
    that LF means a linefeed and CR is ignored (because it was most
    likely added by HOST). However, headers contain CRs that are real
    linefeeds. Note that this might return lines longer than 40
    characters; display fixes those up.

    \param array    data
    \param size     length
    \param rewrap   true iff data is from inbox (and user wants rewrap),
                    false if from outbox

    More or less a copy of SendMsg::PChar2Edit, ReadMsg::DecodeMessage. */
string_t
getMessageFromArray(const char* array, std::size_t size, bool rewrap)
{
    string_t result;
    enum { None, Before, Inside } rewrap_status;

    /* Winplan fixup needed? */
    rewrap_status = None;
    if (rewrap) {
        for (std::size_t i = 0; i < size; ++i) {
            if (array[i] == 10 + 13) {
                rewrap_status = Before;
                break;
            }
        }
    }

    /* Strip trailing blanks */
    while (size && (array[size-1] == 32+13 || array[size-1] == 13+13 || array[size-1] == 13+10))
        --size;

    /* Convert message. */
    bool skipping = false;
    for (std::size_t i = 0; i < size; ++i) {
        char c = array[i] - 13;
        switch (c) {
         case 13:
            if (rewrap_status != Inside) {
                if (rewrap_status == Before
                    && result.size() && result[result.size()-1] == '\n')
                {
                    // a blank line, i.e. end of the headers
                    rewrap_status = Inside;
                }
                result += '\n';
            }
            skipping = false;
            break;
         case 10:
            if (rewrap_status == Inside)
                result += '\n';
            skipping = false;
            break;
         case 0:
            skipping = true;
            break;
         default:
            if (!skipping)
                result += Recoder::getE2I().recodeCharacter(c);
        }
    }
    return result;
}
