/**
  *  \file game/sim-data.cc
  *  \brief Battle Simulator - Data
  *
  *  (c) 2001-2006 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/sim.h"
#include "game/config.h"

/*! \class GSimObject
    \brief Base class for simulator objects

    Contains the definition and state of an object. In addition to the
    normal specification data, this includes some counters for
    statistics tracking: the simulator updates these counters the same
    way as it updates the shield level. */

/** Constructor. Create blank object. */
GSimObject::GSimObject()
    : id(1), damage(0), shield(100), owner(12), experience_level(0), flags(0),
      flak_r_override(0), flak_c_override(0)
{
    storeBasicString(name, "?");
    storeBasicString(fcode, "???");
}

/** Destructor. */
GSimObject::~GSimObject()
{ }

/** Mark this object changed. */
void
GSimObject::setChanged()
{
    changed = true;
}

/** Mark this object unchanged. */
void
GSimObject::setUnchanged()
{
    changed = false;
}

/** Check whether this object was changed. */
bool
GSimObject::isChanged() const
{
    return changed;
}

/** Get object Id. */
int
GSimObject::getId() const
{
    return id;
}

/** Change object Id. */
void
GSimObject::setId(int id)
{
    this->id = id;
    setChanged();
}

/** Get name of this object. */
string_t
GSimObject::getName() const
{
    return getBasicString(name);
}

/** Change object name. */
void
GSimObject::setName(string_t name)
{
    storeBasicString(this->name, name);
    setChanged();
}

/** Get friendly code. */
string_t
GSimObject::getFCode() const
{
    return getBasicString(fcode);
}

/** Change friendly code. */
void
GSimObject::setFCode(string_t fcode)
{
    storeBasicString(this->fcode, fcode);
    setChanged();
}

/** Get damage level. */
int
GSimObject::getDamage() const
{
    return damage;
}

/** Change damage level. */
void
GSimObject::setDamage(int damage)
{
    this->damage = damage;
    setChanged();
}

/** Get shield level. */
int
GSimObject::getShield() const
{
    return shield;
}

/** Change shield level. */
void
GSimObject::setShield(int shield)
{
    this->shield = shield;
    setChanged();
}

/** Get owner. */
int
GSimObject::getOwner() const
{
    return owner;
}

/** Change owner. */
void
GSimObject::setOwner(int owner)
{
    this->owner = owner;
    setChanged();
}

/** Get experience level. */
int
GSimObject::getExperienceLevel() const
{
    return experience_level;
}

/** Change experience level. */
void
GSimObject::setExperienceLevel(int experience_level)
{
    this->experience_level = experience_level;
    setChanged();
}

/** Get flags (fl_XXX). */
int
GSimObject::getFlags() const
{
    return flags;
}

/** Change flags (fl_XXX). */
void
GSimObject::setFlags(int flags)
{
    this->flags = flags;
    setChanged();
}

/** Get FLAK rating override. */
int32
GSimObject::getFlakRatingOverride() const
{
    return flak_r_override;
}

/** Set FLAK rating override. Must be enabled in flags to be actually
    used for simulation. */
void
GSimObject::setFlakRatingOverride(int32 r)
{
    flak_r_override = r;
    setChanged();
}

/** Get FLAK compensation override. */
int32
GSimObject::getFlakCompensationOverride() const
{
    return flak_c_override;
}

/** Change FLAK compensation override. Must be enabled in flags to be
    actually used for simulation. */
void
GSimObject::setFlakCompensationOverride(int32 r)
{
    flak_c_override = r;
    setChanged();
}

/** Get maximum permitted damage level. Considers owner race. Intended
    for display / editing only, so this may allow more than the host
    allows. */
int
GSimObject::getMaxDamage() const
{
    if (raceId(getOwner()) == 2)
        return 150;
    else
        return 99;
}

/** Get maximum permitted shield level. Considers owner race and
    current damage. Intended for display / editing only, so this may
    allow more than the host allows. */
int
GSimObject::getMaxShield() const
{
    int limit = getMaxDamage() - getDamage() + 1;
    if (limit > 100)
        limit = 100;
    return limit;
}

/** Get this object as a ship.
    \return \c this if this is a ship, 0 otherwise. */
GSimShip*
GSimObject::asShip()
{
    return 0;
}

/** Get this object as a planet.
    \return \c this if this is a planet, 0 otherwise. */
GSimPlanet*
GSimObject::asPlanet()
{
    return 0;
}

/******************************** GSimShip *******************************/

/*! \class GSimShip
    \brief Ship Data for Simulator */

/** Construct blank simulator ship. */
GSimShip::GSimShip()
    : crew(10), hull(0), mass(100), beam_type(0), beam_count(0),
      torp_type(0), torp_lcount(0), bay_count(0), ammo(0),
      engine_type(NUM_ENGINES), aggressiveness(agg_Passive),
      intercept_id(0)
{ }

/** Construct from file data.
    \param file_data [in] Ship record read from a .ccb file
    \param version   [in] .ccb version (last digit of file magic) */
GSimShip::GSimShip(const TSimShipData& file_data, int version)
    : crew(file_data.vcro.crew), hull(file_data.hull_type),
      mass(file_data.mass), beam_type(file_data.vcro.beam_type),
      beam_count(file_data.vcro.beam_count), torp_type(file_data.vcro.torp_type),
      torp_lcount(file_data.vcro._torp_launcher_count),
      bay_count(file_data.vcro.bay_count), ammo(file_data.vcro._ammo),
      engine_type(file_data.engine_type), aggressiveness(file_data.aggressiveness),
      intercept_id(file_data.xa_id)
{
    setExperienceLevel(file_data.vcro.experience_level);
    setOwner(file_data.vcro.owner);
    setId(file_data.vcro.id);
    setDamage(file_data.vcro.damage);
    setShield(file_data.shield);
    setName(getBasicString(file_data.vcro.name));
    setFCode(getBasicString(file_data.fcode));
    setFlags(file_data.flags);
    setFlakRatingOverride(file_data.flak_r);
    setFlakCompensationOverride(file_data.flak_c);
    if (version <= 0)
        setAggressiveness(agg_Kill);
    if (version <= 2)
        setFlags(0);
    if (version <= 3) {
        setFlags(getFlags() & ~fl_RatingOverride);
        setInterceptId(0);
    }
}

GSimShip::~GSimShip()
{ }

/** Store data into external representation.
    \param file_data [out] Data
    \param version   [in] Version to store for. Optimize/adapt output for
                     this file format version. */
void
GSimShip::storeData(TSimShipData& file_data, int version)
{
    (void) version;

    storeBasicString(file_data.vcro.name, getName());
    file_data.vcro.damage               = getDamage();
    file_data.vcro.crew                 = getCrew();
    file_data.vcro.id                   = getId();
    file_data.vcro.owner                = getOwner();
    file_data.vcro.race_or_zero         = 0;   // we don't use this field
    file_data.vcro.picture              = 0;   // we don't use this field
    file_data.vcro.hull_or_zero         = 0;   // we don't use this field
    file_data.vcro.beam_type            = getBeamType();
    file_data.vcro.beam_count           = getBeamCount();
    file_data.vcro.experience_level     = getExperienceLevel();
    file_data.vcro.bay_count            = getBayCount();
    file_data.vcro.torp_type            = getTorpType();
    file_data.vcro._ammo                = getAmmo();
    file_data.vcro._torp_launcher_count = getTorpLauncherCount();
    file_data.engine_type               = getEngineType();
    file_data.hull_type                 = getHull();
    file_data.shield                    = getShield();
    storeBasicString(file_data.fcode, getFCode());
    file_data.aggressiveness            = getAggressiveness();
    file_data.mass                      = getMass();
    file_data.flags                     = getFlags();
    file_data.flak_r                    = getFlakRatingOverride();
    file_data.flak_c                    = getFlakCompensationOverride();
    file_data.xa_id                     = getInterceptId();
}

/** Get crew. */
int
GSimShip::getCrew() const
{
    return crew;
}

/** Change crew. */
void
GSimShip::setCrew(int crew)
{
    this->crew = crew;
    setChanged();
}

/** Get hull type. Can be 0 for custom ships. */
int
GSimShip::getHull() const
{
    return hull;
}

/** Set hull type. Can be 0 for custom ships. If this changes the hull
    type, it will also update the arms and crew to reflect that new
    type. */
void
GSimShip::setHull(int hull)
{
    if (hull != this->hull) {
        this->hull = hull;
        if (hull) {
            /* it's a real hull. Fill in defaults. */
            const GHull& def = ::getHull(hull);

            // beams
            this->beam_count = def.getMaxBeams();
            if (this->beam_count) {
                if (!this->beam_type)
                    this->beam_type = NUM_BEAMS;
            }

            // torps
            if (def.getNumBays()) {
                this->torp_lcount = this->torp_type = 0;
                this->bay_count = def.getNumBays();
                this->ammo = def.getCargo();
            } else if (def.getMaxLaunchers()) {
                this->torp_lcount = def.getMaxLaunchers();
                if (!this->torp_type)
                    this->torp_type = NUM_TORPS;
                this->ammo = def.getCargo();
                this->bay_count = 0;
            } else {
                this->torp_lcount = this->torp_type = 0;
                this->bay_count = 0;
                this->ammo = 0;
            }

            // rest
            this->crew = def.getCrew();
            this->mass = def.getMass();
        }
        setChanged();
    }
}

/** Get mass. Only valid for custom ships. */
int
GSimShip::getMass() const
{
    return mass;
}

/** Change mass. Only used for custom ships. */
void
GSimShip::setMass(int mass)
{
    this->mass = mass;
    setChanged();
}

/** Get beam type. */
int
GSimShip::getBeamType() const
{
    return beam_type;
}

/** Change beam type. */
void
GSimShip::setBeamType(int beam_type)
{
    this->beam_type = beam_type;
    setChanged();
}

/** Get beam count. */
int
GSimShip::getBeamCount() const
{
    return beam_count;
}

/** Change beam count. */
void
GSimShip::setBeamCount(int beam_count)
{
    this->beam_count = beam_count;
    setChanged();
}

/** Get torpedo type. */
int
GSimShip::getTorpType() const
{
    return torp_type;
}

/** Change torpedo type. */
void
GSimShip::setTorpType(int torp_type)
{
    this->torp_type = torp_type;
    setChanged();
}

/** Get torpedo launcher count. */
int
GSimShip::getTorpLauncherCount() const
{
    return torp_lcount;
}

/** Change torpedo launcher count. */
void
GSimShip::setTorpLauncherCount(int torp_lcount)
{
    this->torp_lcount = torp_lcount;
    setChanged();
}

/** Get number of fighter bays. */
int
GSimShip::getBayCount() const
{
    return bay_count;
}

/** Change number of fighter bays. */
void
GSimShip::setBayCount(int bay_count)
{
    this->bay_count = bay_count;
    setChanged();
}

/** Get ammunition (torps, fighters). */
int
GSimShip::getAmmo() const
{
    return ammo;
}

/** Change ammunition (torps, fighters). */
void
GSimShip::setAmmo(int ammo)
{
    this->ammo = ammo;
    setChanged();
}

/** Get engine type. */
int
GSimShip::getEngineType() const
{
    return engine_type;
}

/** Change engine type. */
void
GSimShip::setEngineType(int engine_type)
{
    this->engine_type = engine_type;
    setChanged();
}

/** Get aggressiveness. */
int
GSimShip::getAggressiveness() const
{
    return aggressiveness;
}

/** Change aggressiveness (agg_XXX). */
void
GSimShip::setAggressiveness(int aggressiveness)
{
    this->aggressiveness = aggressiveness;
    setChanged();
}

/** Get Id for intercept-attack. */
int
GSimShip::getInterceptId() const
{
    return intercept_id;
}

/** Set Id for intercept-attack. */
void
GSimShip::setInterceptId(int id)
{
    this->intercept_id = id;
    setChanged();
}

GSimShip*
GSimShip::asShip()
{
    return this;
}

/** Check whether this is a custom ship. Custom ships have some more
    freedom in configuring. */
bool
GSimShip::isCustomShip() const
{
    return getHull() == 0;
}

/** Check whether this ship matches the current ship list. */
bool
GSimShip::isMatchingShipList() const
{
    /* custom ships have full freedom */
    if (isCustomShip())
        return false;

    /* valid hull? */
    /* FIXME: we cannot handle these during simulation so we should
       avoid even loading them. */
    if (getHull() > getNumHulls())
        return false;
    const GHull& h = ::getHull(getHull());

    /* beams */
    if (getBeamCount() > h.getMaxBeams())
        return false;

    /* torps / fighters */
    if (getTorpLauncherCount() > h.getMaxLaunchers())
        return false;
    if (getBayCount() != h.getNumBays())
        return false;
    if (getAmmo() > h.getCargo())
        return false;

    return true;
}

/******************************* GSimPlanet ******************************/

/*! \class GSimPlanet
    \brief Planet Data for Simulator */

/** Constructor. Construct blank planet. */
GSimPlanet::GSimPlanet()
    : defense(10), base_defense(10), beam_tech(0),
      torp_tech(1), base_fighters(0)
{
    for (int i = 0; i < NUM_TORPS; ++i)
        base_torps[i] = 0;
}

/** Construct from file data.
    \param file_data [in] Record read from .ccb file
    \param version   [in] File version */
GSimPlanet::GSimPlanet(const TSimPlanetData& file_data, int version)
    : defense(file_data.defense),
      base_defense(file_data.base_defense),
      beam_tech(file_data.beam_tech),
      torp_tech(file_data.torp_tech),
      base_fighters(file_data.base_fighters)
{
    for (int i = 0; i < NUM_TORPS; ++i)
        base_torps[i] = 0;
    if (beam_tech < 0 || beam_tech > NUM_BEAMS)
        beam_tech = 0;
    if (torp_tech <= 0 || torp_tech > NUM_TORPS)
        torp_tech = 1;
    if (base_defense < 0 || base_defense > 200 /* FIXME */)
        base_defense = 0;
    if (defense < 0 || defense > 1000 /* FIXME */)
        defense = 0;

    // setExperienceLevel(file_data.experience_level); // FIXME: struct is wrong
    setOwner(file_data.owner);
    setId(file_data.id);
    setDamage(0); // FIXME?
    setShield(100); // FIXME?
    setName("Planet"); /* FIXME: use spec */
    setFCode(getBasicString(file_data.fcode));
    setFlags(file_data.flags);
    setFlakRatingOverride(file_data.flak_r);
    setFlakCompensationOverride(file_data.flak_c);

    if (version > 1) {
        for (int i = 0; i < NUM_TORPS; ++i)
            base_torps[i] = file_data.base_torps[i];
    } else {
        base_torps[torp_tech-1] = file_data.base_torps_old;
    }
    if (version <= 2)
        setFlags(0);
    if (version <= 3)
        setFlags(getFlags() & ~fl_RatingOverride);
}

/** Destructor. */
GSimPlanet::~GSimPlanet()
{ }

/** Store data into external representation.
    \param file_data [out] Data
    \param version   [in] Version to store for. Optimize/adapt output for
                     this file format version. */
void
GSimPlanet::storeData(TSimPlanetData& file_data, int version)
{
    if (version > 1) {
        for (int i = 0; i < NUM_TORPS; ++i)
            file_data.base_torps[i] = base_torps[i];
        file_data.base_torps_old = 0;
    } else {
        for (int i = 0; i < NUM_TORPS; ++i)
            file_data.base_torps[i] = 0;
        file_data.base_torps_old = (torp_tech > 0 && torp_tech <= NUM_TORPS ? base_torps[torp_tech-1] : 0);
    }

    file_data._pad0            = 0;
    file_data.id               = getId();
    file_data.owner            = getOwner();
    file_data._pad1            = 0;
    file_data.beam_tech        = getBaseBeamTech();
    file_data._pad2            = 0;
    file_data.experience_level = getExperienceLevel();
    file_data.base_fighters    = getBaseFighters();
    file_data._pad3            = 0;
    file_data.torp_tech        = getBaseTorpTech();
    file_data.base_defense     = getBaseDefense();
    file_data.defense          = getDefense();
    file_data.shield           = getShield();
    storeBasicString(file_data.fcode, getFCode());
    file_data.aggressiveness   = -1 /* getAggressiveness() */;
    file_data._pad5            = 0;
    file_data.flags            = getFlags();
    file_data.flak_r           = getFlakRatingOverride();
    file_data.flak_c           = getFlakCompensationOverride();
    file_data._pad6            = 0;
}

/** Get number of planetary defense posts. */
int
GSimPlanet::getDefense() const
{
    return defense;
}

/** Change number of planetary defense posts. */
void
GSimPlanet::setDefense(int defense)
{
    this->defense = defense;
    setChanged();
}

/** Get number of base defense posts. */
int
GSimPlanet::getBaseDefense() const
{
    return base_defense;
}

/** Change number of base defense posts. */
void
GSimPlanet::setBaseDefense(int base_defense)
{
    this->base_defense = base_defense;
    setChanged();
}

/** Get starbase beam tech level. Zero means no base. */
int
GSimPlanet::getBaseBeamTech() const
{
    return beam_tech;
}

/** Change starbase beam tech level. Zero means no base. */
void
GSimPlanet::setBaseBeamTech(int beam_tech)
{
    this->beam_tech = beam_tech;
    setChanged();
}

/** Get starbase torpedo tech level. */
int
GSimPlanet::getBaseTorpTech() const
{
    return torp_tech;
}

/** Change starbase torpedo tech level. */
void
GSimPlanet::setBaseTorpTech(int torp_tech)
{
    this->torp_tech = torp_tech;
    setChanged();
}

/** Get number of starbase fighters. */
int
GSimPlanet::getBaseFighters() const
{
    return base_fighters;
}

/** Change number of starbase fighters. */
void
GSimPlanet::setBaseFighters(int base_fighters)
{
    this->base_fighters = base_fighters;
    setChanged();
}

/** Get number of starbase torps of a given type. */
int
GSimPlanet::getBaseTorps(int tech) const
{
    ASSERT(tech > 0 && tech <= NUM_TORPS);
    return base_torps[tech-1];
}

/** Change number of starbase torps of a given type. */
void
GSimPlanet::setBaseTorps(int tech, int amount)
{
    ASSERT(tech > 0 && tech <= NUM_TORPS);
    this->base_torps[tech-1] = amount;
    setChanged();
}

/** Get number of base torpedoes for a given tech. Computes the
    effective torpedo count from all storage torpedoes. */
int
GSimPlanet::getBaseTorpsAsType(int tech) const
{
    uint32 total_cost = 0;
    for (int i = 1; i <= NUM_TORPS; ++i)
        total_cost += getBaseTorps(i) * getTorp(i).getTorpCost();

    uint32 cost = getTorp(tech).getTorpCost();
    if (cost != 0)
        total_cost /= cost;
    // FIXME: some range checking here?
    return total_cost;
}

GSimPlanet*
GSimPlanet::asPlanet()
{
    return this;
}
