/**
  *  \file game/sim-if.h
  */
#ifndef PCC_V2_GAME_SIM_IF_H
#define PCC_V2_GAME_SIM_IF_H

#include "game/sim.h"

/** Interface between Simulator and Game. Provides methods for the
    simulator setup code to adapt to various environments. When we're
    running from a full client with game data available, we want to
    allow data transfers to and from the game. We don't want to depend
    on the full game data, though, to be able to run standalone. */
class GSimulatorGameInterface {
 public:
    /** Relation of a ship to the simulation. */
    enum ShipRelation {
        ShipIsUnknown,          ///< This is not the same ship as in the simulation.
        ShipIsPlayed,           ///< This is a ship we play. We can modify it.
        ShipIsForeign           ///< This is a foreign ship. We can only watch it.
    };

    GSimulatorGameInterface()
        { }
    virtual ~GSimulatorGameInterface()
        { }

    /** Get name of a planet. */
    virtual string_t getPlanetName(int id) = 0;

    /** Get real owner of a game ship.
        \return real ship owner, 0 if unknown */
    virtual int getShipOwner(int id) = 0;

    /** Check whether game data is available. This decides
        availability of the game interface buttons. */
    virtual bool isGameAvailable() = 0;

    /** Update ship with data from game. */
    virtual bool updateFromGame(GSimShip& sh) = 0;

    /** Update game data with this ship. */
    virtual bool updateToGame(const GSimShip& sh) = 0;

    /** Check whether this planet is the same as the one in the
        game. */
    virtual bool isGamePlanet(const GSimPlanet& pl) = 0;

    /** Update planet with data from game. */
    virtual bool updateFromGame(GSimPlanet& pl) = 0;

    /** Update ship with data from game. */
    virtual bool updateToGame(const GSimPlanet& pl) = 0;

    /** Check whether this ship is a game ship. */
    virtual ShipRelation getShipRelation(const GSimShip& sh) = 0;

    /** Select ship for user. \return true on success. */
    virtual bool selectGameShip(const GSimShip& sh) = 0;
};

class GSimulatorNullInterface : public GSimulatorGameInterface {
 public:
    string_t getPlanetName(int id);
    int getShipOwner(int id);
    bool isGameAvailable();
    bool updateFromGame(GSimShip& sh);
    bool updateToGame(const GSimShip& sh);
    bool isGamePlanet(const GSimPlanet& pl);
    bool updateFromGame(GSimPlanet& pl);
    bool updateToGame(const GSimPlanet& pl);
    ShipRelation getShipRelation(const GSimShip& sh);
    bool selectGameShip(const GSimShip& sh);
};

#endif
