/*
   reading .vpm files

   This reads a .vpm file and outputs a ship summary to standard output.

   Not all values are 100% exact because the movie recorder doesn't
   record everything correctly and uniquely:
   - carriers may be shown to have torps, if the ship previously in that
     slot had some
   - the turn numbers when the ship was built might be guestimated wrong
 */

#include <iostream>
#include <cstdlib>
#include "io/streams.h"
#include "cpluslib/buffer.h"
#include "util/except.h"
#include "game/struct.h"
#include "cpluslib/types.h"
#include "cpluslib/format.h"
#include "cpluslib/misc.h"
#include "game/global.h"
#include "game/spec.h"
#include "tests/movieread.h"

struct Ship {
    uint8   hull;
    uint8   owner_engine;
    uint8   beams;
    uint8   torps;
    uint8   damage;
    uint16  x, y;
    int     built;
};

struct Mine {
    uint16  x, y, radius;
    uint8   owner_type;
};

static int  planets[NUM_PLANETS];
static Ship ships[NUM_SHIPS];
static Mine mines[500];

static int& p(int x) {
    static int dummy;
    if (x > 0 && x <= NUM_PLANETS)
        return planets[x-1];
    else
        return dummy;
}

static Ship& s(int x) {
    static Ship dummy;
    if (x > 0 && x <= NUM_SHIPS)
        return ships[x-1];
    else
        return dummy;
}

static Mine& m(int x) {
    static Mine dummy;
    if (x > 0 && x <= countof(mines))
        return mines[x-1];
    else
        return dummy;
}

class MyMovieReader : public MovieSimpleReader {
 public:
    void changeMinefield(int id, int x, int y, int radius, int ot);
    void changeMinefield(int id, int radius);
    void changePlanet(int id, int owner, bool base);
    void changeShip(int id, int what, int hull, int oe,
                    int beam, int torp, int damage, int x, int y);
};

void
MyMovieReader::changeMinefield(int id, int x, int y, int radius, int ot)
{
    Mine& m = ::m(id);

    m.x          = x;
    m.y          = y;
    m.radius     = radius;
    m.owner_type = ot;
    
}

void
MyMovieReader::changeMinefield(int id, int radius)
{
    m(id).radius = radius;
}

void
MyMovieReader::changePlanet(int id, int owner, bool base)
{
    p(id) = owner | (base << 5);
}

void
MyMovieReader::changeShip(int id, int what, int hull, int oe,
                          int beam, int torp, int damage, int x, int y)
{
    Ship& s = ::s(id);

    if (what & si_OwnerEngine) {
        s.owner_engine = oe;
        if ((oe & 15) == 0)
            s.beams = s.torps = s.hull = 0;
    }
    if (what & si_Hull)
        s.hull = hull, s.built = getTurn();
    if (what & si_Beam)
        s.beams = beam;
    if (what & si_Torp)
        s.torps = torp;
    if (what & si_Damage)
        s.damage = damage;
    if (what & si_X)
        s.x = x;
    if (what & si_Y)
        s.y = y;
}

void
printSummary(int pid)
{
    std::cout << "Summary for player " << pid
              << "\n======================\n\n";
    std::cout << "Ship#  Type                            Beams  Torps     X,Y     Built\n"
              << "-----  ------------------------------  -----  -----  ---------  -----\n";
    for (int sid = 1; sid <= NUM_SHIPS; ++sid) {
        Ship& s = ::s(sid);
        if ((s.owner_engine & 15) == pid) {
            std::cout << format("%4d   %-30s  ", sid, getHull(s.hull).getName());
            if (s.beams > 15)
                std::cout << format("%2dx%-2d  ", s.beams >> 4, s.beams & 15);
            else
                std::cout << "       ";
            if (s.torps > 15)
                std::cout << format("%2dx%-2d  ", s.torps >> 4, s.torps & 15);
            else
                std::cout << "       ";
            std::cout << format("%4d,%4d  %4d\n", s.x, s.y, s.built);
        }
    }
    std::cout << "\n\n";
}

int main(int argc, char** argv)
{
    try {
        if (argc != 2 && argc != 3) {
            std::cerr << "usage: " << argv[0] << " filename.vpm [gamedir]\n";
            std::exit(1);
        }

        initGameGlobals(argc == 3 ? argv[2] : ".", ".");
        initSpecification();

        Ptr<Stream> s = Stream::openFile(argv[1], Stream::C_READ);
        MyMovieReader m;
        m.loadFromStream(*s);
        m.read();
        for (int i = 1; i <= NUM_PLAYERS; ++i)
            printSummary(i);
        std::cout << "Notes: information is not 100% exact. The part of the movie recorder which\n"
                     "records ship equipment is seems not to be very well-tested (the player does\n"
                     "not display that information). The `Built' column is an estimate when this\n"
                     "ship was built. Weapon information is in format `count x type index'.\n";
    }
    catch(std::exception& e) {
        std::cerr << "exception: " << e.what() << "\n";
    }
}
