/**
  *  \file ui/msgbox.cc
  *  \brief Implementation of Standard Message Boxes
  *
  *  (c) 2001-2006 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/msgbox.h"
#include "gfx/font.h"
#include "ui/util.h"
#include "ui/window.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"
#include "ui/widgets.h"
#include "ui/layout.h"
#include "ui/commands.h"
#include "util/nls.h"

UIMultilineStatic::UIMultilineStatic(int font, string_t text, int width, int id)
    : UIWidget(id), font(font), text(text), width(width),
      lines(countLines(fonts[font], width, text))
{
    setState(st_Disabled, true);
}

void
UIMultilineStatic::draw(GfxCanvas& can, UpdateInfo& ui)
{
    GfxParameters parm;
    parm.color = standard_colors[COLOR_BLACK];
    parm.font = fonts[font];
    outTextFormatted(can, parm, extent.x, extent.y, extent.w, text);
    ui.update(getExtent());
}

void
UIMultilineStatic::getLayoutInfo(LayoutInfo& li)
{
    li.flags = 0;
    li.min_size = li.pref_size = GfxPoint(width, font_heights[font] * lines);
}

/******************************* WDialogBox ******************************/

UIMessageBox::UIMessageBox(string_t text, string_t title)
    : UIWindow(title, UIWindow::BLUE_WINDOW /* FIXME? */, COLOR_GRAY, 10),
      group(h.add(new UIGroup(UIHBoxLayout::instance5))),
      layout10(10)
{
    setLayout(layout10);

    int width = fonts[FONT_TITLE]->getTextWidth(text);
    if (width < 200)
        width = 200;
    else if (width > 440)
        width = 400;            // no typo

    add(h.add(new UIMultilineStatic(FONT_TITLE, text, width)));
    add(group);

    group.add(h.add(new UISpacer()));
}

UIMessageBox&
UIMessageBox::addButton(int id, string_t btn, UIKey k)
{
    group.add(h.add(new UIButton(id, btn, k, UIButton::aStop)));
    return *this;
}

bool
UIMessageBox::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key && second_pass) {
        if (e.key.code == SDLK_ESCAPE) {
            stop(cm_Escape);
            return true;
        }
        if (e.key.code == SDLK_RETURN) {
            stop(cm_Enter);
            return true;
        }
    }
    return UIWindow::handleEvent(e, second_pass);
}

int
UIMessageBox::run()
{
    group.add(h.add(new UISpacer()));
    pack();

    ui_root->centerWidget(*this);
    ui_root->add(*this);
    int rv = ui_root->run();
    ui_root->close(*this);
    return rv;
}

bool
UIMessageBox::doYesNoDialog()
{
    UIKeyString y(_("Yes")), n(_("No"));
    addButton(cm_Enter,  y.getString(), y.getKey());
    addButton(cm_Escape, n.getString(), n.getKey());
    return run() == cm_Enter;
}

void
UIMessageBox::doOkDialog()
{
    addButton(cm_Enter, _("OK"), ' ');
    run();
}

/** Standard Message Box. A message box containing the specified text
    and title, and an "OK" button. The message box is modal. The text
    can contain '\n' to start new lines; it is automatically
    word-wrapped. */
void
messageBox(string_t text, string_t title)
{
    return UIMessageBox(text, title).doOkDialog();
}
